import numpy as np
from pyPhasesML.datapipes.DataPipe import DataPipe


class BatchMap(DataPipe):
    def __init__(
        self,
        datapipe: DataPipe,
        batchSize=1,
        onlyFullBatches=True,
        toXY=True,
        toNumpy=True,
    ) -> None:
        super().__init__(datapipe)
        self.batchSize = batchSize
        self.onlyFullBatches = onlyFullBatches
        self.toXY = toXY
        self.toNumpy = toNumpy

    def __getitem__(self, index):
        start_idx = index * self.batchSize
        end_idx = (index + 1) * self.batchSize
        batchX, batchY = [], []

        for i in range(start_idx, min(end_idx, len(self.datapipe))):
            X, Y = self.datapipe[i]
            batchX.append(X)
            batchY.append(Y)

        if self.toNumpy:
            batchX, batchY = np.array(batchX), np.array(batchY)

        return batchX, batchY if self.toXY else zip(batchX, batchY)

    def __len__(self):
        if self.onlyFullBatches:
            return len(self.datapipe) // self.batchSize
        else:
            return (len(self.datapipe) + self.batchSize - 1) // self.batchSize



from dataclasses import astuple, dataclass

import numpy as np


class ValidDataset:
    pass


@dataclass
class DataSet:
    x: np.array
    y: np.array
    numClasses: int = None
    catMatrix = None

    def __len__(self):
        return 2

    @staticmethod
    def fromTuple(data, numClasses=None):
        x, y = data
        return DataSet(x, y, numClasses)

    def asTuple(self):
        return astuple(self)

    def __getitem__(self, index):
        if index == 0:
            return self.x
        if index == 1:
            return self.y
        if index == 2:
            raise StopIteration
        raise TypeError("Datset can only be split in x and y, try to get index %i" % (index))


class TrainingSetLoader:
    def __init__(self, trainingData=None, validationData=None):
        self.trainingData: DataSet = trainingData
        self.validationData: DataSet = validationData



from pyPhases import classLogger
from pyPhasesML.adapter.torch.Callback import Callback

from pyPhases import CSVLogger as Logger


@classLogger
class CSVLogger(Callback):
    def getCsvPath(self):
        return f"{self.config.logPath}/{self.config.csvLogFile}"

    def onTrainingStart(self, model, dataset):
        self.logger = Logger(self.getCsvPath())

    def onValidationEnd(self, model, results, scorer):
        csvRow = {
            "epoch": model.epoch,
        }

        csvRow.update(model.runningStats)
        metricsValues = {m: results[m] for m in scorer.metrics}

        for metricName in metricsValues:
            csvRow[f"val_{metricName}"] = metricsValues[metricName]

        self.logger.addCsvRow(csvRow)



from typing import List

import numpy as np

from pyPhases import classLogger
from pyPhasesRecordloader import ChannelsNotPresent, Event, RecordSignal, Signal


class PreprocessSteMissingException(Exception):
    pass


@classLogger
class SignalPreprocessing:
    def __init__(self, config) -> None:
        self.config = config
        self.stepsPerType = config["stepsPerType"]

    def step(self, stepname, signal: Signal, recordSignal: RecordSignal, **options):
        # check if augmentation step exist

        if hasattr(self, stepname):
            # call method
            return getattr(self, stepname)(signal, recordSignal, **options)
        else:
            raise PreprocessSteMissingException(f"Preprocessing {stepname} not found")

    def preprocessingSignal(self, psgSignal: RecordSignal):
        self.preprocessSignalByType(psgSignal, self.stepsPerType)

    def preprocessSignalByType(self, psgSignal: RecordSignal, stepsPerType):
        for signal in psgSignal.signals:
            cName = signal.name
            if cName in psgSignal.signalNames:
                signal = psgSignal.getSignalByName(cName)
                signalType = signal.typeStr

                if signalType in stepsPerType:
                    stepNames = stepsPerType[signalType]

                    for processStep in stepNames:
                        options = {}
                        if isinstance(processStep, dict):
                            name = processStep["name"]
                            options = {key: value for key, value in processStep.items() if key != "name"}
                            processStep = name
                        self.step(processStep, signal, psgSignal, **options)
                elif signalType is not None:
                    self.logError(
                        f"Signaltype {signal.type} for signal {signal.name} has no preprocessing steps (defined in preprocessing.stepsPerType.[type])"
                    )
            elif cName not in self.optionalSignals and not signal["generated"]:
                self.logError(f"Missing channel {cName} for {signal.recordId}")
                raise ChannelsNotPresent(cName, signal.recordId)

    def extendEvents(self, events: List[Event], extendEventsByName: dict):
        if extendEventsByName is not None:
            for event in events:
                if event.name in extendEventsByName:
                    addBefore, addAfter = extendEventsByName[event.name]
                    event.start -= addBefore
                    event.duration += addBefore + addAfter
        return events

    def resampleSimple(self, signal: Signal, recordSignal: RecordSignal, targetFrequency=None):
        targetFrequency = targetFrequency or recordSignal.targetFrequency
        factor = signal.frequency / targetFrequency
        if factor >= 1:
            signal.signal = signal.signal[:: int(factor)]
        else:
            signal.signal = np.repeat(signal.signal, int(1 / factor), axis=0)
        signal.frequency = targetFrequency

    def standardisation(self, signal: Signal, recordSignal: RecordSignal):
        center = np.mean(signal.signal)
        scale = np.std(signal.signal)
        if scale == 0:
            scale = 1
        signal.signal = (signal.signal - center) / scale

    def cut(self, signal: Signal, recordSignal: RecordSignal, minValue, maxValue):
        signal.signal[signal.signal > maxValue] = maxValue
        signal.signal[signal.signal < minValue] = minValue

    def normalize(self, signal: Signal, recordSignal: RecordSignal, minValue=0, maxValue=1):
        signalMin = min(signal.signal)
        signalMax = max(signal.signal)
        if minValue is None:
            minValue = signalMin

        if maxValue is None:
            maxValue = signalMax

        signalMinAbs = -1 * signalMin
        signal.signal = (signal.signal + signalMinAbs) / (signalMinAbs + signalMax)
        signal.signal *= maxValue - minValue
        signal.signal += minValue

    def normalizePercentage(self, signal: Signal, recordSignal: RecordSignal):
        self.cut(signal, recordSignal, 0, 100)
        self.normalize(signal, recordSignal, 0, 1)

    def resample(self, signal: Signal, recordSignal: RecordSignal, targetFrequency=None):
        from scipy.signal import resample
        targetFrequency = targetFrequency or recordSignal.targetFrequency

        signalLength = int(len(signal.signal) * targetFrequency / signal.frequency)
        signal.signal = resample(signal.signal, signalLength)
        signal.frequency = targetFrequency



import math
from pathlib import Path
from pyPhases import classLogger
from pyPhasesML.Model import ModelConfig
from pyPhasesML.adapter.torch.Callback import Callback


@classLogger
class CyclicLearningrate(Callback):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self.learningRateMax = config.learningRate
        self.learningRateMin = config.learningRate / config.cycleLRDivisor
        self.stepSizeInEpochs = 4

    def onTrainingStart(self, model, dataset):
        import torch.optim as optim
        def cyclical_lr(stepsize, min_lr=3e-4, max_lr=3e-3):
            # Scaler: we can adapt this if we do not want the triangular CLR
            scaler = lambda x: 1.0

            # calculate the LR
            def getLR(it):
                return min_lr + (max_lr - min_lr) * relative(it, stepsize)

            # Additional function to see where on the cycle we are
            def relative(it, stepsize):
                cycle = math.floor(1 + it / (2 * stepsize))
                x = abs(it / stepsize - 2 * cycle + 1)
                return max(0, (1 - x)) * scaler(cycle)

            return getLR

        # setup optimizer and scheduler
        model.optimizer = optim.Adam(model.model.parameters(), lr=1.0)
        step_size = self.stepSizeInEpochs * len(dataset.trainingData)
        clr = cyclical_lr(step_size, min_lr=self.learningRateMin, max_lr=self.learningRateMax)
        model.batchScheduler = optim.lr_scheduler.LambdaLR(model.optimizer, [clr])

    def getResumeBatchOptimizerPath(self):
        return Path(self.getLogPath(), ".resumeBatchScheduler.pt")

    def onCheckpoint(self, model):
        import torch
        torch.save(model.batchScheduler.state_dict(), str(self.getResumeBatchOptimizerPath()))
    
    def onRestore(self, model):
        import torch
        batchSchedulerState = torch.load(str(self.getResumeBatchOptimizerPath()))
        model.batchScheduler.load_state_dict(batchSchedulerState)



from .RecordMap import RecordMap


class FoldMapper:
    def __init__(self, datasetLength, folds, validationPostion):
        self.folds = folds
        self.validationPostion = validationPostion

        records = range(datasetLength)
        validataionSize = len(records) // self.folds
        if validataionSize == 0:
            raise Exception("The dataset is too small for the given number of folds")
        self.validationRecords = records[
            self.validationPostion * validataionSize : (self.validationPostion + 1) * validataionSize
        ]
        self.trainingsRecords = [record for record in records if record not in self.validationRecords]

    def getTraining(self, dataset):
        return RecordMap(dataset, self.trainingsRecords)

    def getValidation(self, dataset):
        return RecordMap(dataset, self.validationRecords)

    def getTrainingMap(self):
        return self.trainingsRecords

    def getValidationMap(self):
        return self.validationRecords



from pyPhasesML.datapipes.DataPipe import DataPipe


class RecordMap(DataPipe):
    def __init__(self, dataset: DataPipe, mapping):
        self.dataset = dataset
        self.mapping = mapping
        self.currentIndex = 0

    def __len__(self):
        return len(self.mapping)

    def __getitem__(self, index):
        return self.dataset[self.mapping[index]]




from pyPhasesML.datapipes import DataPipe

from ..DataAugmentation import DataAugmentation


class Augmentor:
    def __init__(self, datapipe: DataPipe, augmentation: DataAugmentation, config) -> None:
        self.datapipe = datapipe
        self.augmentation = augmentation
        self.config = config

    def __getitem__(self, index):
        return self.augmentation(self.datapipe[index], self.config, index)

    def __len__(self):
        return len(self.datapipe)




from queue import Queue, Full
from threading import Condition, Lock, Thread, current_thread

from pyPhasesML.datapipes.DataPipe import DataPipe
from pyPhases import classLogger


@classLogger
class PreloadPipe(DataPipe):
    def __init__(self, datapipe: DataPipe, preloadCount=1):
        super().__init__(datapipe)
        self.queue = Queue(maxsize=preloadCount)
        self.done = False

    def start(self):
        thread = Thread(target=self._preload)
        thread.daemon = True  # don't wait for thread to finish
        thread.start()

    def _preload(self):
        for d in self.datapipe:
            self.queue.put(d, block=True)
        self.done = True
        return

    def __iter__(self):
        self.start()
        self.done = False
        return self

    def __next__(self):
        if self.done and self.queue.empty() or len(self) == 0:
            raise StopIteration

        return self.queue.get(block=True)




class DataPipe:
    def __init__(self, datapipe: "DataPipe") -> None:
        self.datapipe = datapipe

    def __getitem__(self, index):
        return self.datapipe[index]

    def __len__(self):
        return len(self.datapipe)
    
    def __iter__(self) -> int:
        for i in range(len(self)):
            yield self[i]

    def close(self):
        pass


import torch.optim as optim
from pyPhases import classLogger
from pyPhasesML.Model import ModelConfig
from pyPhasesML.adapter.torch.Callback import Callback


class OptimizerNotFound(Exception):
    pass


@classLogger
class LoadOptimizer(Callback):
    def __init__(self, modelConfig: ModelConfig) -> None:
        super().__init__(modelConfig)
        self.optimizerId = modelConfig.optimizerId
        self.lr = modelConfig.learningRate
        self.lr_decay = modelConfig.learningRateDecay

    def onTrainingStart(self, model, dataset):
        if self.optimizerId == "adams":
            model.optimizer = optim.Adam(model.model.parameters(), lr=self.lr, weight_decay=self.lr_decay)
        elif self.optimizerId == "sgd":
            model.optimizer = optim.SGD(model.model.parameters(), lr=self.lr, weight_decay=self.lr_decay)
        elif self.optimizerId == "nesterov":
            model.optimizer = optim.SGD(model.model.parameters(), lr=self.lr, weight_decay=self.lr_decay, nesterov=True)
        elif self.optimizerId == "nadams":
            model.optimizer = optim.NAdam(model.model.parameters(), lr=self.lr, momentum_decay=self.lr_decay)
        else:
            raise OptimizerNotFound(f"optimizer {self.optimizerId} is currently not supported for pytorch implementation")



from pyPhasesML.Model import ModelConfig


class Callback:
    def __init__(self, config: ModelConfig, priority=100) -> None:
        self.config = config
        self.priority = priority

    def getLogPath(self):
        return f"{self.config.logPath}/"

    def trigger(self, event, *args, **kwargs):
        if event == "trainingStart":
            self.onTrainingStart(*args, **kwargs)
        elif event == "trainingEnd":
            self.onTrainingEnd(*args, **kwargs)
        elif event == "validationStart":
            self.onValidationStart(*args, **kwargs)
        elif event == "validationEnd":
            self.onValidationEnd(*args, **kwargs)
        elif event == "batchEnd":  # at the end of a batch, after backpropagation and batch scheduler
            self.onBatchEnd(*args, **kwargs)
        elif event == "shutdown":  # when a shutdown request is made
            self.onShutdown(*args, **kwargs)
        elif event == "checkpoint":  # when a checkpoint is created
            self.onCheckpoint(*args, **kwargs)
        elif event == "restore":  # when a restore checkpoint was found and the training is restored
            self.onRestore(*args, **kwargs)
        elif event == "register":  # when the callback gets registered to the model
            self.onRegister(*args, **kwargs)

    def onTrainingStart(self, model, dataset):
        pass

    def onTrainingEnd(self, model):
        pass

    def onValidationStart(self, model, validationData):
        pass

    def onValidationEnd(self, model, results, scorer):
        pass

    def onBatchEnd(self, model, batchIndex):
        pass
    
    def onShutdown(self, model):
        pass
    
    def onCheckpoint(self, model):
        pass

    def onRestore(self, model):
        pass
    
    def onRegister(self, model):
        pass



from pathlib import Path
import pickle
import timeit

from pyPhases import classLogger

from pyPhasesML.adapter.torch.Callback import Callback
from pyPhasesML.Model import ModelConfig


@classLogger
class CheckPoint(Callback):
    def __init__(self, config: ModelConfig) -> None:
        super().__init__(config)
        self.metricDefinitions = None
        self.notImprovedSince = 0

    def onTrainingStart(self, model, dataset):
        self.epochStartTime = timeit.default_timer()

    def onValidationStart(self, model, dataset):
        self.epochEndTime = timeit.default_timer()

    def prettyPrintConfusionMatrix(self, confusion_matrix):
        num_rows, num_cols = confusion_matrix.shape
        max_value_length = max(len(str(confusion_matrix.max())), len(str(confusion_matrix.min())))
        separator = "-" * ((max_value_length + 2) * num_cols)

        rows = []
        for i in range(num_rows):
            row_data = [f"{int(confusion_matrix[i, j]):>{max_value_length}}" for j in range(num_cols)]
            rows.append(" | ".join(row_data))

        print(separator)
        print("\n".join(rows))
        print(separator)

    def onValidationEnd(self, model, results, scorer):
        metricsValues = {m: results[m] for m in scorer.metrics}
        metricDiffStrings = []
        metricValuetrings = []
        improved = False
        modelId = "checkpointModel_%i_" % model.epoch

        if self.metricDefinitions is None:
            self.metricDefinitions = {m: scorer.getMetricDefinition(m) for m in model.validationMetrics}

        globalBestMetric = model.validationMetrics[0]

        metricStrings = []
        for metricName, metricVal in metricsValues.items():
            bestValue, useAsBest, biggerIsBetter = self.metricDefinitions[metricName]
            diff = metricVal - bestValue
            metricStrings.append(f"{metricName}: " + "{:.3f}".format(metricVal) + " [best: {:.3f}]".format(bestValue))
            metricDiffStrings.append(f"{metricName}: " + "{:.3f}".format(diff))
            metricValuetrings.append("{:.3f}".format(metricVal))

            isBigger = metricVal > bestValue

            if (biggerIsBetter and isBigger) or (not biggerIsBetter and not isBigger):
                self.metricDefinitions[metricName][0] = metricVal
                if useAsBest:
                    improved = True
                    if metricName == globalBestMetric:
                        model.bestMetric = max(model.bestMetric, metricsValues[globalBestMetric])

        validationEndTime = timeit.default_timer()

        epochTime = str(self.epochEndTime - self.epochStartTime)
        validationTime = str(validationEndTime - self.epochStartTime)
        self.log(f"Validation-Epoch Number: {str(model.epoch)}  Epoche Time: {epochTime}  Validation Time: {validationTime}")
        if model.useGPU:
            import torch
            self.log(f"Max GPU memory usage: {torch.cuda.max_memory_allocated() / (1024 ** 3):.2f} GB")

        if "confusion" in scorer.results:
            self.prettyPrintConfusionMatrix(scorer.results["confusion"])

        trainingStats = " | ".join([f"{n}:{v}" for n, v in model.runningStats.items()])
        self.log(f"Training Stats: {trainingStats} ")
        self.log(" ".join(metricStrings))

        if improved:
            self.log("Model Improved: " + " ".join(metricDiffStrings))
            path = f"{self.getLogPath()}/{modelId}" + "_".join(metricValuetrings) + ".pkl"
            with open(path, "wb") as f:
                import torch
                torch.save(model.model.state_dict(), f)
            self.notImprovedSince = 0
            model.bestModelPath = path
        else:
            self.notImprovedSince += 1
            self.log("Model not improving since %i epochs" % (self.notImprovedSince))

        if self.config.stopAfterNotImproving > 0 and self.notImprovedSince >= self.config.stopAfterNotImproving:
            model.finish = True

        model.metricDefinitions = self.metricDefinitions

    def getCheckpointFile(self):
        return Path(self.getLogPath(), ".resumeCheckpoint.pt")

    def onCheckpoint(self, model):

        with open(self.getCheckpointFile(), "wb") as f:
            pickle.dump({
                "notImprovedSince": self.notImprovedSince,
                "metricDefinitions": self.metricDefinitions,
                "bestModelPath": model.bestModelPath,
            }, f, protocol=pickle.HIGHEST_PROTOCOL)
    
    def onRestore(self, model):
        if not self.getCheckpointFile().exists():
            return

        with open(self.getCheckpointFile(), "rb") as f:
            state = pickle.load(f)
            self.notImprovedSince = state["notImprovedSince"]
            self.metricDefinitions = state["metricDefinitions"]
            model.metricDefinitions = state["metricDefinitions"]
            model.bestModelPath = state["bestModelPath"]



from .DataPipe import DataPipe
from .BatchMap import BatchMap
from .ShuffleMap import ShuffleMap
from .PreloadPipe import PreloadPipe


class DataLoader:
    @staticmethod
    def build(
        data: DataPipe,
        batchSize=1,
        onlyFullBatches=False,
        preload=True,
        shuffle=False,
        shuffleSeed=None,
        mapBatchToXY=True,
        torch=False,
        torchCuda=False,
    ):
        if shuffle:
            data = ShuffleMap(data, seed=shuffleSeed)

        if batchSize > 1:
            data = BatchMap(data, batchSize, onlyFullBatches, toXY=mapBatchToXY)

        if torch:
            from .TorchMap import TorchMap

            data = TorchMap(data, cuda=torchCuda)

        if preload:
            data = PreloadPipe(data)

        return data
